use lib '../mgmt/perl';

#################################################################################
# Created       : Martin Foster
# Modified      : 24-Aug-2006
#################################################################################
#
# Configure - Commandline based script that is used for Ethereal Realms that will
#             allow for the addition, removal and management of databases used
#             in single and virtual-domain based operations of the site.
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################
use Config::IniFiles;						# Configuration files
use DBI;							# Database indepedent
use Digest::MD5 qw(md5_hex);					# MD5 hash
use File::Copy;							# Simplified file copies
use strict;							# Variable locality

#################################################################################
# Data Members
#################################################################################

my $conf;							# Configuration handle
my $data;							# Database handle 
my $lang;							# Language handle

my $capture;							# Captured text
my $count;							# Site counter

#################################################################################
# Program Area
#################################################################################

	# Make sure everything is fine
	if (Initialize(\$conf, \$lang, \$data))
	{
		# Loop until needed
		do
		{
			# Pull amount of sites
			($count) = Get($data, "SELECT COUNT(*)
				FROM public.Site");


			# Basic check
			if (defined($capture))
			{
				# Detailed options
				# Site Creation
				if ($capture =~ /^C/i)
				{
					# Creation 
					Create($data, $conf, $lang, $count);
				}

				# Additional control
				elsif ($count > 0)
				{
					# Site Removal
					if (($capture =~ /^D/i)
					 && ($count > 1))
					{
						# Removal of entries
						Delete($data, $conf, $lang);
					}

					# Site Modification
					elsif ($capture =~ /^M/i)
					{
						# Modification of entries
						Change($data, $conf, $lang);
					}

					# Generation of templates
					elsif ($capture =~ /^G/i)
					{
						# Create a new release file
						Release($data, $conf, $lang);
					}

					# Upgrading of sites
					elsif (($capture =~ /^U/i)
					 && (-f 'upgrade.tmpl'))
					{
						# Upgrade to a new version
						Prepare($data, $conf, $lang);
					}
				}
			}


			# Clear screen
			system "clear";

			# Menu generation
			# Title bar
			print $lang->val('Notices','NoticeMenu'), "\n\n";

			# Initial setup
			if ($count < 1)
			{
				# Simply Section and menu
				print "  ",   $lang->val('Messages','MenuSecSite'), "\n";
				print "    ", $lang->val('Messages','MenuCreate'), "\n\n";
			}

			# Basic design
			else
			{
				# Site management section
				print "  ",   $lang->val('Messages','MenuSecSite'), "\n";
				print "    ", $lang->val('Messages','MenuCreate'), "\n";
				print "    ", $lang->val('Messages','MenuUpgrade'), "\n" if (-f 'upgrade.tmpl');
				print "    ", $lang->val('Messages','MenuDelete'), "\n";
				print "    ", $lang->val('Messages','MenuModify'), "\n\n";

				# Database management
				print "  ",   $lang->val('Messages','MenuSecData'), "\n";
				print "    ", $lang->val('Messages','MenuRelease'), "\n\n";
			}

			# Capture information
			$capture = Capture("  " . $lang->val('Messages','MenuPrompt'));


		} while ($capture !~ /^Q/i);
	}


	# Safety
	$data->disconnect() if (defined($data));

#################################################################################
# Sub-Routines
#################################################################################

#####################
# Capture Input
#
# User input is critical to many parts of the configuration process, as a
# result it becomes necessary to make sure that some attempts are made to have
# things work properly.
#
# This script attempts to address this by making sure defaults are selected if
# the user does not input any information.  If defaults are lacking it will
# refuse blank answers completely as to ensure a degree of sanity.

sub Capture
{
	#####################
	# Data members
	my $question	= shift;				# Question asked
	my $default	= shift;				# Default value

	my $capture;						# Captured line
	my $display;						# Displayed default
	
	#####################
	# Program area

	# Always assigned
	while (!defined($capture))
	{
		# Ensure definition
		$display = (defined($default)) ? $default : '';

		# Prompt human for question
		print $question . " [" . $display . "] ";


		# Await human input
		$capture = <STDIN>;


		# Assign default if need be
		$capture = ($capture =~ /^\n/) ? $default : $capture;
	}
	
	# Carriage return
	chomp($capture);

	# Return
	return $capture;
}


#####################
# Change
#
# Change is a sub-routine that is designed to modify sites, allowing one
# to enable, disable and even set the defaults.   Essentially this allows
# for a more granular control of sites.

sub Change
{
	#####################
	# Data members
	my $data      = shift;					# Database handle
	my $conf      = shift;					# Configuration file
	my $lang      = shift;					# Language file

	my $alias;						# Aliases
	my $capture;						# Captured information
	my $count;						# Individual counter
	my $size;						# Number of sites
	
	my @aliases;						# List of aliases	
	my @sites;						# List of sites

	#####################
	# Program area

	# Pull list of sites
	List($data, \@sites, "SELECT 
		 SiteSchema
		FROM public.Site
		ORDER BY SiteSchema");

	# Determine count
	$size = @sites;	

	
	# Clear screen
	system "clear";

	# Display notice
	print $lang->val('Notices','NoticeChange'), "\n\n";

	# Loop and display
	for ($count=1; $count <= $size; $count++)
	{
		# Individual sites
		print "  \[$count\] " . $sites[$count-1] . "\n";
	}

	# Spacer
	print "\n\n";
	
	# Prompt
	$capture = Capture("  " . $lang->val('Messages','SiteModify'), 0);

	# Valid selection
	if (($capture =~ /^\d+$/)
	 && ($capture > 0)
	 && ($capture <= $size))
	{
		# Pull information
		my ($schema, $name, $domain, $root, $auth, $shared, $active, $default) = Get($data, "SELECT
			 SiteSchema,
			 SiteName,
			 SiteDomain,
			 SiteRoot,
			 SiteAuth,
			 SiteShared,
			 SiteActive,
			 SiteDefault
			FROM public.Site
			WHERE SiteSchema=?", $sites[$capture-1]
		 );
		
		# Confirmatory data
		print "\n  "   . $lang->val('Messages','SiteSecConfirm'), "\n";
		print "    " . $lang->val('Messages','SiteName')    . ": " . $name    . "\n";
		print "    " . $lang->val('Messages','SiteSchema')  . ": " . $schema  . "\n";
		print "    " . $lang->val('Messages','SiteDomain')  . ": " . $domain  . "\n";
		print "    " . $lang->val('Messages','SiteRoot')    . ": " . $root    . "\n";
		print "    " . $lang->val('Messages','SiteAuth')    . ": " . $auth    . "\n";
		print "    " . $lang->val('Messages','SiteShared')  . ": " . $shared  . "\n";
		print "    " . $lang->val('Messages','SiteActive')  . ": " . $active  . "\n";
		print "    " . $lang->val('Messages','SiteDefault') . ": " . $default . "\n\n\n";

		# Confirm changeable options
		$name    = Capture("    "    . $lang->val('Messages','SiteName'),    $name);
		$domain  = lc(Capture("    " . $lang->val('Messages','SiteDomain'),  $domain));
		$root    = lc(Capture("    " . $lang->val('Messages','SiteRoot'),    $root));
		$auth    = lc(Capture("    " . $lang->val('Messages','SiteAuth'),    $auth));
		$shared  = lc(Capture("    " . $lang->val('Messages','SiteShared'),  $shared));
		$active  = lc(Capture("    " . $lang->val('Messages','SiteActive'),  $active));
		$default = lc(Capture("    " . $lang->val('Messages','SiteDefault'), $default));

		# Ensure values
		$default = ($default =~ /^y/i) ? 'yes' : 'no'; 
		$active  = ($active  =~ /^y/i) ? 'yes' : 'no'; 

		# Sanity in auth
		$auth    = md5_hex($name)          if (length($auth) != 32);

		# Sanity in shared
		$shared  = lc(substr($name, 0, 4)) if (length($shared) != 4);
		
		# Update information
		eval
		{
			# Update information
			$data->do("UPDATE public.Site
				SET SiteName=?,
				    SiteDomain=?,
				    SiteRoot=?,
				    SiteAuth=?,
				    SiteShared=?,
				    SiteActive=?,
				    SiteDefault=?
				WHERE SiteSchema=?", {},

			 $name,
			 $domain,
			 $root,
			 $auth,
			 $shared,
			 $active,
			 $default,
			 $schema
			);

			# Commit changes
			$data->commit();
		};

		# Error took place
		if ($@)
		{
			# Rollback changes
			$data->rollback();

			# Display error
			die "TRANSACTION ERROR: $@";
		}


		# Aliases
		# Debugging
		print "\n\n  "   . $lang->val('Messages','SiteSecAlias'), "\n";

		# Pull list of sites
		List($data, \@aliases, "SELECT 
			 AliasAddress
			FROM public.Alias
			WHERE SiteSchema=?
			ORDER BY AliasAddress", $schema);


		# Loop and confirm
		# Existence of sites
		foreach $alias (@aliases)
		{
			# Update information
			$capture = Capture("    " . $lang->val('Messages','SiteAliasMaintain') . " " . $alias, 'yes');


			# Remove information
			if ($capture =~ /^n/i)
			{
				# Evaluate statement
				eval
				{
					# Update information
					$data->do("DELETE FROM public.Alias 
						WHERE SiteSchema=?
						  AND AliasAddress=?", {},

					 $schema,
					 $alias
					);

					# Commit changes
					$data->commit();
				};

				# Error took place
				if ($@)
				{
					# Rollback changes
					$data->rollback();

					# Display error
					die "TRANSACTION ERROR: $@";
				}
			}
		}


		# Create a new entry
		do
		{
			# Update information
			$capture = Capture("    " . $lang->val('Messages','SiteAliasCreate'), 'done');

			# Worth a peak
			if ($capture !~ /^done/i)
			{
				# Evaluate statement
				eval
				{
					# Update information
					$data->do("INSERT INTO public.Alias
						(SiteSchema,
						 AliasAddress)
						VALUES(?,?)", {},

					 $schema,
					 $capture
					);

					# Commit changes
					$data->commit();
				};

				# Error took place
				if ($@)
				{
					# Rollback changes
					$data->rollback();

					# Display error
					die "TRANSACTION ERROR: $@";
				}
			}
		} while ($capture !~ /^done/i);

		# Display completion
		Capture("\n  " . $lang->val('Messages','GeneralDone') . " ", 'Enter');
	}
}


#####################
# Create
#
# Create allows the administrator to create new sites that can be accessed
# using the scripts.    This is a necessary step in the use of the system
# as the base scripts will only configure the necessary groundwork to be
# able to track and create sites.

sub Create
{
	#####################
	# Data members
	my $data      = shift;					# Database handle
	my $conf      = shift;					# Configuration file
	my $lang      = shift;					# Language file
	my $count     = shift;					# Number of instances

	my $capture;						# Captured information
	
	my $name;						# Sites Full Name
	my $schema;						# Site short name/schema name
	my $domain;						# Domain name
	my $root;						# Root path of http files
	my $auth;						# Authentication password
	my $shared;						# Shared memory access
	my $active;						# Active status of site
	my $default;						# Default status						

	#####################
	# Program area

	# Clear screen
	system "clear";

	# Display notice
	print $lang->val('Notices','NoticeCreate'), "\n\n";

	# Begin prompting questions
	# Basics
	$name   = Capture("    " . $lang->val('Messages','SiteName'));
	$schema = lc(Capture("    " . $lang->val('Messages','SiteSchema')));
	$domain = lc(Capture("    " . $lang->val('Messages','SiteDomain')));
	$root   = lc(Capture("    " . $lang->val('Messages','SiteRoot')));

	# Auth
	$auth   = lc(Capture("    " . $lang->val('Messages','SiteAuth'), md5_hex($name)));

	# Shared memory
	$shared = lc(Capture("    " . $lang->val('Messages','SiteShared'), lc(substr($name, 0, 4))));

	# Presets available
	$active = lc(Capture("    " . $lang->val('Messages','SiteActive'), 'yes'));

	# Need for default?
	if ($count > 0)
	{
		# Other sites exist
		$default = lc(Capture("    " . $lang->val('Messages','SiteDefault'), 'no'));
	}
	else
	{
		# Forcefully assign default
		$default = 'yes';
	}

	# Ensure values
	$default = ($default =~ /^y/i) ? 'yes' : 'no'; 
	$active  = ($active  =~ /^y/i) ? 'yes' : 'no'; 

	
	# Spacer
	print "\n     * * \n\n";

	# Confirmatory data
	print "    " . $lang->val('Messages','SiteName')    . ": " . $name    . "\n";
	print "    " . $lang->val('Messages','SiteSchema')  . ": " . $schema  . "\n";
	print "    " . $lang->val('Messages','SiteDomain')  . ": " . $domain  . "\n";
	print "    " . $lang->val('Messages','SiteRoot')    . ": " . $root    . "\n";
	print "    " . $lang->val('Messages','SiteAuth')    . ": " . $auth    . "\n";
	print "    " . $lang->val('Messages','SiteShared')  . ": " . $shared  . "\n";
	print "    " . $lang->val('Messages','SiteActive')  . ": " . $active  . "\n";
	print "    " . $lang->val('Messages','SiteDefault') . ": " . $default . "\n\n";

	# Confirm
	$capture = Capture("  " . $lang->val('Messages','SiteConfirm'), 'yes');	

	# Confirmed
	# Begin update
	if ($capture =~ /^y/i)
	{
		# Safety
		eval
		{
			# Create initial entry
			# Site
			$data->do("INSERT INTO public.Site
				(SiteSchema,
				 SiteName,
				 SiteDomain,
				 SiteRoot,
				 SiteAuth,
				 SiteShared,
				 SiteActive,
				 SiteDefault)
				VALUES(?,?,?,?,?,?,?,?)", {},

			 $schema,
			 $name,
			 $domain,
			 $root,
			 $auth,
			 $shared,
			 $active,
			 $default);

		 	# Initial alias
			$data->do("INSERT INTO public.Alias
				(SiteSchema,
				 AliasAddress)
				VALUES(?,?)", {},

			 $schema,
			 $domain);


		 	# Replication
			# Schema
			Generate($data, $conf, $lang, $conf->val('Settings','TemplateSchema'),    $schema);

			# Schema functions
			Generate($data, $conf, $lang, $conf->val('Settings','TemplateFunctions'), $schema);

			# Schema views
			Generate($data, $conf, $lang, $conf->val('Settings','TemplateViews'),     $schema);

			# Individual data
			Generate($data, $conf, $lang, $conf->val('Settings','TemplateData'),      $schema);

			# Display notice
			print $lang->val('Notice','NoticeGenerate'), "\n";

			# Display command to use
			print '> cat ' . $conf->val('Script','ScirptWorkLoc') . '/' . $schema . '.sql |' .
			      ' psql -h ' . $conf->val('Database','PgHost') .
			           ' -p ' . $conf->val('Database','PgPort') .
			           ' -U ' . $conf->val('Admin','AdmUser'), "\n\n";

			# Commit changes
			$data->commit();
		};

		if ($@)
		{
			# Rollback changes
			$data->rollback();

			# Display error
			die "TRANSACTION ERROR: $@";
		}
	}

	# Display completion
	Capture("\n  " . $lang->val('Messages','GeneralDone') . " ", 'Enter');
}


#####################
# Delete
#
# Will remove sites as requested after confirming the administrators wish.  Works
# by showing users a list of all sites and allowing a user to have it removed by
# selecting the sequence number.

sub Delete
{
	#####################
	# Data members
	my $data      = shift;					# Database handle
	my $conf      = shift;					# Configuration file
	my $lang      = shift;					# Language file

	my $capture;						# Captured information
	my $confirm;						# Confirmatory
	my $count;						# Individual counter
	my $size;						# Number of sites
	
	my @sites;						# List of sites

	#####################
	# Program area

	# Pull list of sites
	List($data, \@sites, "SELECT 
		 SiteSchema
		FROM public.Site
		ORDER BY SiteSchema");

	# Determine count
	$size = @sites;	

	
	# Clear screen
	system "clear";

	# Display notice
	print $lang->val('Notices','NoticeDelete'), "\n\n";

	# Loop and display
	for ($count=1; $count <= $size; $count++)
	{
		# Individual sites
		print "  \[$count\] " . $sites[$count-1] . "\n";
	}

	# Spacer
	print "\n\n";
	
	# Prompt
	$capture = Capture("  " . $lang->val('Messages','SiteRemove'), 0);

	# Valid selection
	if (($capture =~ /^\d+$/)
	 && ($capture > 0)
	 && ($capture <= $size))
	{
		# Confirm
		$confirm = Capture("  " . $lang->val('Messages','SiteConfirm'), 'no');

		# Confirmed
		if ($confirm =~ /^y/i)
		{
			# Evaluate
			eval
			{
				# Debugging
				print "  " . $lang->val('Messages','SiteRemoving');
				
				# Remove site entry
				$data->do("DELETE FROM public.Site WHERE SiteSchema=?", {}, $sites[$capture-1]);

				# Debugging
				print ".";
				 
				# Remove schema
				$data->do("DROP SCHEMA " . $sites[$capture-1] . " CASCADE", {});

				# Debugging
				print ".";

				# Commit changes
				$data->commit();

				# Display completion
				Capture("  " . $lang->val('Messages','GeneralDone') . "  ", 'Enter');
			};

			# Failed
			if ($@)
			{
				# Rollback changes
				$data->rollback();

				# Kill entry
				die("TRANSACTION ERROR: $@");
			}
		}
	}
}

	
#####################
# Data Format
#
# Will make sure to return things properly for use in the display.   Otherwise
# things may not work as expected.

sub Format
{
	#####################
	# Data members

	my $data    = shift;					# Piece to check 


	#####################
	# Program area

	# Defined
	if (defined($data))
	{
		# Format cleanly
		$data =~ s/(\'|\"|\`)/\\$1/gs;
		$data =~ s/\n/\\n/gs;
		$data =~ s/\r/\\r/gs;
		$data =~ s/\\/\\/gs;

		# Release renames
		$data =~ s/(erealms\.org|ethereal-realms\.org)/PGDOMAIN/gsi;
		$data =~ s/(Ethereal Realms)/PGSITE/gsi;
		$data =~ s/(Ethereal)/PGSHORT/gsi;

		# File location
		$data =~ s/\/home\/erealms/PGROOT/gsi;

		# Return
		return "'$data'";
	}

	# Nulls
	else
	{
		# Null value
		return 'NULL';
	}
}


#####################
# Generate
#
# Generate is designed to work around limitations in coding.   This will generate
# the necessary document necessary to completely initialize a site and then have
# pgsql run for the user.
#
# This solves a few problems, one of which is that it will generate a useable
# document to look up in case of troubleshooting and it will make use of the more
# forgiving psql to get things setup right.

sub Generate
{
	#####################
	# Data members

	my $data     = shift;					# Database handle
	my $conf     = shift;					# Configuration handle
	my $lang     = shift;					# Language handle
	my $file     = shift;					# Filename
	my $schema   = shift;					# Schema generation
	my $handle   = shift;					# Filehandle for writing

	my $line;						# Passed line

	my $owner;						# User owning files
	my $passwd;						# Password for user
	my $domain;						# Domain name
	my $site;						# Site name
	my $short;						# Short name
	my $root;						# Root site address


	#####################
	# Program area

	# Clear screen
	system "clear";

	# Filehandle check
	# Safety check
	return 0 unless (-f $file);

	# Make sure things work
	eval
	{
		# Create read filehandle
		open(IN, "<$file")
		 or die("Unable to open $file to read: $!");

		# Create append filehandle
		open(OUT, ">>" . $conf->val('Script','ScriptWorkLoc') . '/' . $schema . '.sql')
		 or die("Unable to open $schema.sql to append: $!") unless (defined($handle));
	};

	# Avoid trouble
	return 0 if ($@);


	# Pull fill-in values
	$owner  = $conf->val('Script','ScriptUser');
	$passwd = $conf->val('Script','ScriptPass');

	# Pull remaing values
	($site, $domain, $root) = Get($data, "SELECT SiteName, SiteDomain, SiteRoot
		FROM public.Site WHERE SiteSchema=?", $schema);

	# Short name
	$short = ucfirst($schema);
	
	# Continue naturally until EOF
	while ($line = <IN>)
	{
		# Altering the natural course
		next if ($line =~ /^\-\-/);		# Comment lines
		next if ($line =~ /^(\n|\r)/);		# Blank lines

		# Search and replace
		# Owner and pass
		$line =~ s/PGOWNER/$owner/gs;
		$line =~ s/PGPASSWD/$passwd/gs;

		# Schema attributes
		$line =~ s/PGSCHEMA/$schema/gs;

		# Domain name, site and root
		$line =~ s/PGDOMAIN/$domain/gs;
		$line =~ s/PGSITE/$site/gs;
		$line =~ s/PGSHORT/$short/gs;
		$line =~ s/PGROOT/$root/gs;

		# Print out line
		if (defined($handle))
		{
			# Make use of referenced filehandle
			print $handle $line;
		}

		else
		{
			# Tried and true method
			print OUT $line;
		}
	}

	# Cleanup
	close(IN);
	close(OUT) unless (defined($handle));

	# Successful
	return 1;
}



#####################
# GetData
#
# Generic sub-routine that will return a single row of data be it in multiple
# amounts or just one field within the row there is a special section for
# font handling if choice is at all defined.

sub Get
{
	#####################
	# Data members
	my $data     = shift;					# Database handle
	my $query    = shift;					# Database query
	my $statement;						# SQL Statement

	my @values;						# Values returned from table


	#####################
	# Program area

	# Prepare and execute
	$statement = $data->prepare($query);
	$statement->execute(@_);

	# Retreive and finish
	@values = $statement->fetchrow();
	$statement->finish();

	# Return retrieved information
	return @values;
}


#####################
# Initialize Config
#
# Designed to ensure that the configuration file does in fact exist on the
# system.   If this is not the case it will prompt the user for the necessary
# information in order to test the database and set down the necessary
# ground work.
#
# If a user wishes to change the configuration, then the file will be removed
# and the user required to enter information once again.

sub Initialize
{
	#####################
	# Data members
	my $cfg		= shift;				# Configuration handle
	my $lng		= shift;				# Language handle
	my $data        = shift;				# Database handle

	my $conf;						# Configuration file
	my $lang;						# Language file
	
	my $capture;						# Captured line
	my $lib;						# Library check
	my $single;						# Singular selected value
	my $param;						# Selected parameter
	
	my @sections    = qw(Admin Database Script);		# Active sections
	my @parameters;						# Parameters in a section
	my @libs;						# Missing libs


	#####################
	# Program area

	# Safety check
	die "Configuration file missing!\n" unless (-f './conf.ini');
	die "Language file missing!\n"      unless (-f './lang.ini');

	
	# Create instances
	$conf = new Config::IniFiles(-file=>'./conf.ini');
	$lang = new Config::IniFiles(-file=>'./lang.ini');

	# Check for need to prompt user
	# No need exists
	if ($conf->val('Settings','InitialRun') =~ /^no/i)
	{
		# Debugging
		print "  " . $lang->val('Messages','InitCheck') . "...";
		
		# Check for synch
		foreach $single (@sections)
		{
			# Pull parameters
			@parameters = $lang->Parameters($single); 

			# Loop and prompt
			foreach $param (@parameters)
			{
				# Safety check
				die "ERROR: Missing parmeter: $param"
			       		unless (defined($conf->val($single, $param)));
			}
		}

		# Debugging
		print "  " . $lang->val('Messages','GeneralDone') . "\n";


		# Confirmation of acceptance
		$capture = Capture("    " . $lang->val('Messages','InitReset'), 'no');

		# Reset desired
		if ($capture =~ /^y/i)
		{
			# Change for next pass
			$conf->setval('Settings','InitialRun','yes');


			# Debugging
			print "  " . $lang->val('Messages','InitWrite') . "...";

			# Rewrite file
			die "Could not write to file!\n" unless ($conf->RewriteConfig());

			# Debugging
			print "  " . $lang->val('Messages','GeneralDone') . "\n\n";
		}	
	}

	# Need exists
	if ($conf->val('Settings','InitialRun') =~ /^yes/i)
	{
		# Clear screen
		system "clear";

		# Debugging
		print "  " . $lang->val('Messages','InitConf') . "\n";
		
		# Run through each section
		foreach $single (@sections)
		{
			# Debugging
			print "    $single:\n";

			# Pull parameters
			@parameters = $conf->Parameters($single); 

			# Loop and prompt
			foreach $param (@parameters)
			{
				# Set value
				$conf->setval($single, $param, Capture(
					"      " . $lang->val($single, $param),
					$conf->val($single, $param))
				 );
			}
		}

		# Change for next pass
		$conf->setval('Settings','InitialRun','no');


		# Debugging
		print "  " . $lang->val('Messages','InitWrite') . "...";

		# Rewrite file
		die "Could not write to file!\n" unless ($conf->RewriteConfig());

		# Debugging
		print "  " . $lang->val('Messages','GeneralDone') . "\n\n";



		# Rewriting of scripts
		# Prompt
		$capture = Capture("  " . $lang->val('Messages','ReWritePrompt'), 'no');

		# Define need
		if ($capture =~ /^y/i)
		{
			# Safety
			if ((-d $conf->val('Script','ScriptLibLoc'))
			 && (-d $conf->val('Script','ScriptWebRoot')))
		 	{
				# Begin rewriting
				Search($conf, $lang);
			}

			else
			{
				# Directories missing skip rewrite
				print "    " . $lang->val('Messages','ReWriteMissing') . "\n\n";

			}
		}
	}

	# Copy handle
	${$cfg} = $conf;
	${$lng} = $lang;


	# Library existence
	# Prompt
	$capture = Capture("  " . $lang->val('Messages','LibraryPrompt'), 'no');

	if ($capture =~ /^y/i)
	{
		# Debugging
		print "\n    " . $lang->val('Messages','LibraryCheck') . "\n";

		# Pull parameters
		@parameters = $conf->Parameters('Library'); 

		# Initialize
		$lib = 0;
	
		# Loop and prompt
		foreach $param (@parameters)
		{
			# Set value
			$single = $conf->val('Library',$param);

			# Debugging
			print "      $single...";
		
			# Test
			eval "use $single;";

			# Warn user
			if ($@)
			{
				# Debugging
				print $lang->val('Messages','GeneralMissing') . "\n";

				# Push
				push(@libs, $single);

				# Increments
				$lib++;
			}

			# Inform
			else
			{
				# Debugging
				print $lang->val('Messages','GeneralFound') . "\n";
			}

		}

		# Debugging help
		if ($lib > 0)
		{
			# Clear screen
			system "clear";

			# Display notice and example
			print "\n" .$lang->val('Notices','NoticeLibrary') . "\n"
			  .   "  \> perl -MCPAN -e shell\n\n"
			  .   "     cpan shell -- CPAN exploration and modules installation (vx.xx)\n"
			  .   "     ReadLine support enabled\n\n";

			# Loop and display
			foreach $lib (@libs)
			{
				# Proper example
				print "     cpan\> install $lib\n";
			}

			# Spacer
			print "   cpan\> exit\n";
		}

		# Debugging
		Capture("  " . $lang->val('Messages','GeneralDone') . "  ", 'Enter');

		# Fail process
		return 0;
	}


	# Database handle
	# Debugging
	print "\n  " . $lang->val('Messages','InitConnect') . "...";
	
	# Workaround
	my $dbname = $conf->val('Database','PgDb');
	my $dbhost = $conf->val('Database','PgHost');
	my $dbport = $conf->val('Database','PgPort');
	my $dbuser = $conf->val('Admin','AdmUser');
	my $dbpass = $conf->val('Admin','AdmPass');
	
		
	# Remote connection
	eval
	{
		# Establish connection
		${$data} = DBI->connect("DBI:Pg:dbname=$dbname;
		 		host=$dbhost;
		 		port=$dbport;
				user=$dbuser;
				password=$dbpass" 
		 ) || die ("REMOTE FAILED...");

		# Success
		print "REMOTE SUCCESS...\n\n";
	};

	# Local connection
	if ($@)
	{
		print "REMOTE FAILED...";
		
		# Try this method
		eval
		{
			# Establish connection
			${$data} = DBI->connect("DBI:Pg:dbname=$dbname;
					user=$dbuser;
					password=$dbpass", 
			 ) || die ("LOCAL FAILED...");

			# Success
			print "LOCAL SUCCESS...\n\n";
		};

		# Ultimate failure
		if ($@)
		{
			# Clear screen
			system "clear";

			# Debugging
			print "  " . $lang->val('Messages','GeneralFail') . "\n\n";

			# Display notice
			print $lang->val('Notices','NoticeDatabase') . "\n\n";

			
			# Die ungracefully
			die ("Database error \#$DBI::err :  $DBI::errstr");
		}
	}

	# Set settings
	${$data}->{AutoCommit} = 0;
	${$data}->{RaiseError} = 1;
	
	
	# Debugging
	print "  " . $lang->val('Notices','NoticeDatabasePM') . "\n";

	# Initialization of base
	if ($conf->val('Settings','InitialBase') =~ /^no/i)
	{
		# Initialize base or fail
		return 0 unless Replicate(${$data}, $conf, $lang, $conf->val('Settings','TemplateBase'), 'public');

		# Update value
		$conf->setval('Settings','InitialBase','yes');

		# Update changes
		$conf->RewriteConfig();
	}

	# Debugging
	Capture("  " . $lang->val('Messages','GeneralDone') . "  ", 'Enter');

	# Continue
	return 1;
}


#####################
# List
#
# Sub-routine created to take a reference to an array and retrieve a list of
# all records within a specific field of information.   This is a near exact
# copy of the getList sub-routine available in the Ethereal::Database module.

sub List
{
	#####################
	# Data members
	my $data     = shift;					# Database handle
	my $array    = shift;					# Reference of array
	my $query    = shift;					# SQL query

	my $statement;						# SQL Statement


	#####################
	# Program area

	# Prepare and execute
	$statement = $data->prepare($query);
	$statement->execute(@_);

	# Cleanup of array
	splice(@{ $array }, 0);

	# Retrieve and place in array
	while ($_ = $statement->fetchrow())
	{ push(@{ $array }, $_); }

        # End statement
        $statement->finish();
}


#####################
# Prepare Upgrade
#
# Upgrades requires that it knows which sites to replicate information
# to.   As a result, this preperation stage will prompt a user to select
# which sites to replicate information to.
#
# Beginning at first with all of them, this will allow a user to remove
# sites as needed through a toggle effect.   This is necessary since
# developing sites such as Ethereal Realms itself will always make changes
# first on one site in order to test the proverbial waters.

sub Prepare
{
	#####################
	# Data members
	my $data      = shift;					# Database handle
	my $conf      = shift;					# Configuration file
	my $lang      = shift;					# Language file

	my $capture;						# Captured information
	my $site;						# Unique site
	
	my @sites;						# List of sites
	my %sites;						# Array depecting selection
	

	#####################
	# Program area

	# Pull list of sites
	List($data, \@sites, "SELECT 
		 SiteSchema
		FROM public.Site
		ORDER BY SiteSchema");

	# Loop through and discover
	foreach $site (@sites)
	{
		# Initialize hash
		$sites{$site} = 'action';
	}
	
	# Determine sites
	do
	{
		# Clear screen
		system "clear";

		# Display notice
		print $lang->val('Notices','NoticePrep'), "\n\n";

		# Ability to action
		if (defined($capture))
		{
			# Site exists toggle settings
			if (defined($sites{$capture}))
			{
				# Toggle
				$sites{$capture} = ($sites{$capture} eq 'action') ? 'pass' : 'action';
			}
		}	
		
		# Display current status
		# Title bar
		print "    Site Name\tStatus\n    -------- --------\n";
		
		# Individual listing
		foreach $site (sort(keys(%sites)))
		{
			# One site at a time nicely
			print "    " . $site . "\t: " . $sites{$site} . "\n";
		}

		# Capture input
		$capture = Capture("\n  " . $lang->val('Messages','SiteToggle'), 'done');
		
	} while ($capture !~ /^done/i);

	# Cleanup
	foreach $site (sort(keys(%sites)))
	{
		# Cleanup
		delete $sites{$site} if ($sites{$site} eq 'pass');
	}
	
	# Re-Assign
	@sites = (sort(keys(%sites)));

	# Generation aspect
	# Open file
	open(UPGRADE, ">" . $conf->val('Script','ScriptWorkLoc') . '/' . 'upgrade.sql')
	 or die("Unable to open update.sql to append: $!");
	
	# Complete upgrade generation
	foreach $site (@sites)
	{
		# Call replication
		Generate($data, $conf, $lang, 'upgrade.tmpl', $site, \*UPGRADE);
	}

	# Close file
	close(UPGRADE);

	# Display notice
	print $lang->val('Notice','NoticeGenerate'), "\n";

	# Display command to use
	print '> cat ' . $conf->val('Script','ScirptWorkLoc') . '/' . 'upgrade.sql |' .
	      ' psql -h ' . $conf->val('Database','PgHost') .
	           ' -p ' . $conf->val('Database','PgPort') .
	           ' -U ' . $conf->val('Admin','AdmUser'), "\n\n";

	# Commit changes
	$data->commit();
}


#####################
# Pull
#
# Pull is a wrapper for the complexity behind preparing and retreiving
# information using the prepare and execute handlers.

sub Pull
{
	#####################
	# Data members
	my $data     = shift;					# Database handle
	my $sth      = shift;					# Query handle
	my $query    = shift;					# Database query


	#####################
	# Program area

	# Prepare query
	${$sth} = $data->prepare($query);

	# Execute
	return ${$sth}->execute(@_);
}


#####################
# Release
#
# Release is designed to create a base template of information required
# for release of the code or creation in new sites.    Without this
# specific ability, all information would have to be generated from
# pre-defined templates which would in fact hinder operations due to
# updates eventually falling out of favour.

sub Release
{
	#####################
	# Data members
	my $data      = shift;					# Database handle
	my $conf      = shift;					# Configuration file
	my $lang      = shift;					# Language file

	my $default;						# Default schema
	my $file;						# Filename
	my $res;						# Result handle
	my $statement;						# Query

	#####################
	# Program area

	# Open file
	# Pull value
	$file = $conf->val('Settings','TemplateData');

	# Create filehandle
	open(FILE, ">$file")
	 or die("Unable to open $file: $!");
	

	# Clear screen
	system "clear";

 	# Debugging
	print $lang->val('Notices','NoticeRelease');
 
	
	# Determine default schema
	($default) = Get($data, "SELECT SiteSchema
		FROM public.Site
		WHERE SiteDefault=?", 'yes');

	#
	#
	# Users
	# Title
	print FILE "--\n-- DEFAULT USERS\n--\n";
	print FILE "INSERT INTO PGSCHEMA.Puppeteer (PuppeteerLogin,PuppeteerPassword,PuppeteerName,PuppeteerEmail,PuppeteerAdminPrivs,PuppeteerSuperPrivs) VALUES('admin','2d/SWwLlVvQEE','Default Administrator','postmaster\@PGDOMAIN','yes','yes');\n";
	print FILE "INSERT INTO PGSCHEMA.Puppeteer (PuppeteerLogin,PuppeteerPassword,PuppeteerName,PuppeteerEmail) VALUES('system','*************','Default User','postmaster\@PGDOMAIN');\n\n";


	#
	#
	# Realm

	# Title
	print FILE "--\n-- DEFAULT REALMS\n--\n";
	print FILE "INSERT INTO PGSCHEMA.Realm (RealmName, RealmShort, RealmPassword, RealmPublic) VALUES ('system','system','*************','hybrid');\n";
	print FILE "INSERT INTO PGSCHEMA.Realm (RealmName, RealmShort, RealmPassword, RealmPublic) VALUES ('banned','banned','*************','hybrid');\n";


	#
	#
	# Realm Designs
	Pull($data, \$statement, "SELECT
		  TagName       AS \"TagName\",
		  RealmName     AS \"RealmName\",
		  DesignContent AS \"DesignContent\"
		FROM $default.RealmDesign
		WHERE RealmName='system'
		ORDER BY TagName, RealmName");

	# Title
	print FILE "--\n-- REALM DESIGN\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.RealmDesign (TagName,RealmName,DesignContent) VALUES(%s,%s,%s);\n",
			Format($res->{'TagName'}),
			Format($res->{'RealmName'}),
			Format($res->{'DesignContent'});
	}

	# End
	$statement->finish();


	#
	#
	# Astrology
	Pull($data, \$statement, "SELECT
		  AstroSign     AS \"AstroSign\",
		  AstroStart    AS \"AstroStart\",
		  AstroEnd      AS \"AstroEnd\"
		FROM $default.Astrology");

	# Title
	print FILE "--\n-- ASTROLOGY\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Astrology (AstroSign,AstroStart,AstroEnd) VALUES(%s,%s,%s);\n",
			Format($res->{'AstroSign'}),
			Format($res->{'AstroStart'}),
			Format($res->{'AstroEnd'});
	}

	# End
	$statement->finish();

	
	#
	#
	# Contacts
	Pull($data, \$statement, "SELECT
		  ContactName      AS \"ContactName\",
		  ContactType      AS \"ContactType\",
		  ContactFilter    AS \"ContactFilter\"
		FROM $default.Contact
		ORDER BY ContactType, ContactName");

	# Title
	print FILE "--\n-- CONTACT\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Contact (ContactName,ContactType,ContactFilter) VALUES(%s,%s,%s);\n",
			Format($res->{'ContactName'}),
			Format($res->{'ContactType'}),
			Format($res->{'ContactFilter'});
	}

	# End
	$statement->finish();


	#
	#
	# Country
	Pull($data, \$statement, "SELECT
		  CountryTLD          AS \"CountryTLD\",
		  CountryName         AS \"CountryName\",
		  CountryFIPS104      AS \"CountryFIPS104\",
		  CountryISO2         AS \"CountryISO2\",
		  CountryISO3         AS \"CountryISO3\",
		  CountryISONo        AS \"CountryISONo\",
		  CountryCapital      AS \"CountryCapital\",
		  CountryRegions      AS \"CountryRegions\",
		  CountryCurrency     AS \"CountryCurrency\",
		  CountryCurrencyCode AS \"CountryCurrentyCode\",
		  CountryPopulation   AS \"CountryPopulation\"
		FROM $default.Country
		ORDER BY CountryTLD, CountryName");

	# Title
	print FILE "--\n-- COUNTRY\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Country (CountryTLD, CountryName, CountryFIPS104, CountryISO2, CountryISO3, CountryISONo, CountryCapital, CountryRegions, CountryCurrency, CountryCurrencyCode, CountryPopulation) VALUES(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s);\n",
			Format($res->{'CountryTLD'}),
			Format($res->{'CountryName'}),
			Format($res->{'CountryFIPS104'}),
			Format($res->{'CountryISO2'}),
			Format($res->{'CountryISO3'}),
			Format($res->{'CountryISONo'}),
			Format($res->{'CountryCapital'}),
			Format($res->{'CountryRegions'}),
			Format($res->{'CountryCurrency'}),
			Format($res->{'CountryCurrencyCode'}),
			Format($res->{'CountryPopulation'});
	}

	# End
	$statement->finish();


	#
	#
	# Exif Tag Information
	Pull($data, \$statement, "SELECT
		  ExifTag       AS \"ExifTag\",
		  ExifMacro     AS \"ExifMacro\"
		FROM $default.Exif
		ORDER BY ExifTag");

	# Title
	print FILE "--\n-- EXIF TAGS\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Exif (ExifTag, ExifMacro) VALUES(%s,%s);\n",
			Format($res->{'ExifTag'}),
			Format($res->{'ExifMacro'});
	}

	# End
	$statement->finish();

	
	#
	#
	# FAQ
	Pull($data, \$statement, "SELECT
		  FaqQuestion      AS \"FaqQuestion\",
		  FaqGroup         AS \"FaqGroup\",
		  FaqAnswer        AS \"FaqAnswer\"
		FROM $default.Faq
		ORDER BY FaqQuestion, FaqAnswer");

	# Title
	print FILE "\n\n--\n-- FAQ\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Faq (FaqQuestion, FaqGroup, FaqAnswer) VALUES(%s,%s,%s);\n",
			Format($res->{'FaqQuestion'}),
			Format($res->{'FaqGroup'}),
			Format($res->{'FaqAnswer'});
	}

	# End
	$statement->finish();


	#
	#
	# Font
	Pull($data, \$statement, "SELECT
		  FontName         AS \"FontName\",
		  FontType         AS \"FontType\",
		  FontCode         AS \"FontCode\"
		FROM $default.Font
		ORDER BY FontType, FontName");

	# Title
	print FILE "\n\n--\n-- FONT\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Font (FontName,FontType,FontCode) VALUES(%s,%s,%s);\n",
			Format($res->{'FontName'}),
			Format($res->{'FontType'}),
			Format($res->{'FontCode'});
	}

	# End
	$statement->finish();


	#
	#
	# Fortunes
	Pull($data, \$statement, "SELECT
		  FortuneType      AS \"FortuneType\",
		  FortuneMsg       AS \"FortuneMsg\"
		FROM $default.Fortune
		WHERE FortuneMsg IS NOT NULL
		ORDER BY FortuneType");

	# Title
	print FILE "\n\n--\n-- FORTUNES\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Fortune (FortuneType,FortuneMsg) VALUES(%s,%s);\n",
			Format($res->{'FortuneType'}),
			Format($res->{'FortuneMsg'});
	}

	# End
	$statement->finish();


	#
	#
	# Genre
	Pull($data, \$statement, "SELECT
		  GenreName        AS \"GenreName\",
		  GenreType        AS \"GenreType\",
		  GenreSub         AS \"GenreSub\",
		  GenreDescription AS \"GenreDescription\"
		FROM $default.Genre
		ORDER BY GenreType, GenreSub, GenreName");

	# Title
	print FILE "\n\n--\n-- GENRE\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Genre (GenreName,GenreType,GenreSub,GenreDescription) VALUES(%s,%s,%s,%s);\n",
			Format($res->{'GenreName'}),
			Format($res->{'GenreType'}),
			Format($res->{'GenreSub'}),
			Format($res->{'GenreDescription'});
	}

	# End
	$statement->finish();


	#
	#
	# Gummies
	Pull($data, \$statement, "SELECT
		  GummyName        AS \"GummyName\",
		  GummyClass       AS \"GummyClass\",
		  GummyPath        AS \"GummyPath\",
		  GummyIdent       AS \"GummyIdent\",
		  GummyMacro       AS \"GummyMacro\"
		FROM $default.Gummy
		WHERE RealmName=?
		  AND GummyIdent IS NULL
		ORDER BY GummyCLass, GummyName", 'system');

	# Title
	print FILE "\n\n--\n-- GUMMY\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Gummy (GummyName,GummyClass,GummyPath,GummyIdent,GummyMacro) VALUES(%s,%s,%s,%s,%s);\n",
			Format($res->{'GummyName'}),
			Format($res->{'GummyClass'}),
			Format($res->{'GummyPath'}),
			Format($res->{'GummyIdent'}),
			Format($res->{'GummyMacro'});
	}

	# End
	$statement->finish();


	#
	#
	# Signs
	Pull($data, \$statement, "SELECT
		  SignAnimal      AS \"SignAnimal\",
		  SignElement     AS \"SignElement\",
		  SignStart       AS \"SignStart\",
		  SignEnd         AS \"SignEnd\"
		FROM $default.Sign");

	# Title
	print FILE "\n\n--\n-- SIGN\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Sign (SignAnimal,SignElement,SignStart,SignEnd) VALUES(%s,%s,%s,%s);\n",
			Format($res->{'SignAnimal'}),
			Format($res->{'SignElement'}),
			Format($res->{'SignStart'}),
			Format($res->{'SignEnd'});
	}

	# End
	$statement->finish();


	#
	#
	# Tags
	Pull($data, \$statement, "SELECT
		  TagName       AS \"TagName\",
		  TagType       AS \"TagType\",
		  TagData       AS \"TagData\"
		FROM $default.Tag
		WHERE NOT (TagName='SetAddressDebug'
		  AND TagType=?)
		ORDER BY TagType, TagName", 'user');

	# Title
	print FILE "\n\n--\n-- TAG\n--\n";

	# Debugging
	print ".";

	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Display
		printf FILE "INSERT INTO PGSCHEMA.Tag (TagName,TagType,TagData) VALUES(%s,%s,%s);\n",
			Format($res->{'TagName'}),
			Format($res->{'TagType'}),
			Format($res->{'TagData'});
	}

	# End
	$statement->finish();

	#
	#
	# Version number setting

	# Title
	print FILE "\n\n--\n-- SET VERSION NUMBER\n--\n";
	print FILE "UPDATE PGSCHEMA.Tag SET TagData='" . $conf->val('Settings','Version') . "' WHERE TagName='SetVersion' AND TagType='system'";

	# Debugging
	Capture("  " . $lang->val('Messages','GeneralDone') . " ", 'Enter');
}


#####################
# Replicate
#
# Replicate will work differently based on the presence of a filename or
# not.   If a filename is provided, the script will read that file extract
# any relevant SQL statements, replace the macros and then run through a 
# parser before being blessed and incorporated to all associated sites.
#
# The other is more direct, working vaguely how 'psql' would work by taking
# human input, waiting for the familiar ; and then reacting accordingly
# by repeating the steps above.

sub Replicate
{
	#####################
	# Data members

	my $data     = shift;					# Database handle
	my $conf     = shift;					# Configuration handle
	my $lang     = shift;					# Language handle
	my $file     = shift;					# Filename

	my $copy;						# Copy to work on
	my $line;						# Passed line
	my $schema;						# Single schema name
	my $statement;						# Captured statement

	my @schemas  = @_;					# Schemas to handle

	#####################
	# Program area

	# Clear screen
	system "clear";

	# Method of input
	# Filehandle
	if (defined($file))
	{
		# Safety check
		return 0 unless (-f $file);

		# Make sure things work
		eval
		{
			# Create filehandle
			open(FILE, "<$file")
			 or die("Unable to open $file: $!");
		};

		# Avoid trouble
		return 0 if ($@);
	}

	else
	{
		# Display notice
		print $lang->val('Notices','NoticeReplicate') . "\n\n";

		# Copy filehandle
		*FILE = *STDIN;
	}

	# Initialize
	$statement = '';

	# Initial prompt
	print '> ' unless (defined($file));
	
	# Evaluate
	eval
	{
		# Pull fill-in values
		my $owner  = $conf->val('Script','ScriptUser');
		my $passwd = $conf->val('Script','ScriptPass');
		
		# Continue naturally until EOF
		while ($line = <FILE>)
		{
			# Altering the natural course
			last if ($line =~ /^\\q/i);		# End called
			next if ($line =~ /^\-\-/);		# Comment lines
			next if ($line =~ /^(\n|\r)/);		# Blank lines

			# End of input
			if ($line =~ /^\s*(ALTER|CREATE|DROP|INSERT|MODIFY|UPDATE)/si)
			{
				# Search and replace
				$statement =~ s/PGOWNER/$owner/gs;
				$statement =~ s/PGPASSWD/$passwd/gs;

				# Loop and replicate
				foreach $schema (@schemas)
				{
					# Create copy
					$copy = $statement;
					
					# Later input file
					$copy =~ s/PGSCHEMA/$schema/gs;

					# Run operation on copy
					$data->do($copy);
				}

				# Reset statement
				$statement = '';
			}
		
			# Add to statement
			$statement = $statement . $line;

			# Additional prompt
			print '> ' unless (defined($file));
		}

		# Empty buffer
		# Deemed sufficient amount
		if (length($statement) > 15)
		{
			# Search and replace
			$statement =~ s/PGOWNER/$owner/gs;
			$statement =~ s/PGPASSWD/$passwd/gs;

			# Loop and replicate
			foreach $schema (@schemas)
			{
				# Create copy
				$copy = $statement;
					
				# Later input file
				$copy =~ s/PGSCHEMA/$schema/gs;

				# Run operation on copy
				$data->do($copy);
			}
		}
		
		# Commit changes
		$data->commit();
	};

	# Generated error
	if ($@)
	{
		# Display notice
		print $lang->val('Notices','NoticeCommit') . "\n";
		print $statement, "\n" if (defined($file));
		print $@, "\n\n";

		# Rollback changes
		$data->rollback();

		# Cleanup
		close(FILE) if (defined($file));

		# Failed operation
		return 0;
	}

	# Cleanup
	close(FILE) if (defined($file));

	# Successful
	return 1;
}


#####################
# Rewrite
#
# Will change a file if necessary, the first step being that it will verify
# that the appropriate header exists and then change the file as needed.

sub ReWrite
{
	#####################
	# Data members

	my $conf     = shift;					# Configuration handle
	my $lang     = shift;					# Language handle
	my $file     = shift;					# Filename

	my $count    = 0;					# Counter
	my $head;						# Header
	my $line;						# Singlular line


	#####################
	# Program area

	# Header file
	$head = $conf->val('Settings','InitialHeader');
	
	# Open input file handle
	open(INPUT, "<$file")
	 or die ("Unable to open file $file for reading: $!");


	# Loop through file
	while ($line = <INPUT>)
	{
		# Initial check
		if ($count < 1)
		{
			# Confirm the appropriate file exists
			unless ($line =~ /^$head/si)
			{
				# Display Notice
				print "        " . $lang->val('Messages','ReWriteSkip') . " " . $file . "\n";	

				# Close file
				close(INPUT);

				# Exit loop
				last;
			}

			# Prepare for rewrite
			else
			{
				# Display change
				print "        " . $lang->val('Messages','ReWriteModify') . " " . $file;	

				# Open output file handle
				open(OUTPUT, ">$file.new")
				 or die ("Unable to open file $file for writing: $!");

				# Print initial lines
				print OUTPUT "#!" . $conf->val('Script','ScriptPerlLoc') . "\n";
				print OUTPUT "use lib \'" . $conf->val('Script','ScriptLibLoc') . "\';\n";


				# Additional check
				# Already using libraries
				$line = <INPUT>;

				# Re-add only if necessary
				print OUTPUT $line if ($line !~ /^use lib/si);
			}
		}

		# Remaining lines
		else
		{
			# Add in additional lines
			print OUTPUT $line;
		}

		# Increment
		$count++;
	}

	# Post input
	if ($count > 0)
	{
		# Close files
		close(INPUT);
		close(OUTPUT);

		# Status display
		print ".";

		# Remove old file
		unlink("$file")
		  or die("Unable to remove orginal file $file: $!");


		# Status display
		print ".";

		# Rename file
		rename("$file.new", "$file");


		# Status display
		print ".";
		Capture("  " . $lang->val('Messages','GeneralDone') . " ", 'Enter');
	}
}


#####################
# Direcory Search
#
# Will look at a given directory and see what is available to be modified.  In
# addition, it will recursively call sub-directories as to handle those as well.

sub Search
{
	#####################
	# Data members

	my $conf      = shift;					# Configuration file
	my $lang      = shift;					# Language file
	my $dir       = shift;					# Directory location
	my $file;						# File name

	my @locales;						# Additional directories

	#####################
	# Program area

	# Safety
	$dir = (defined($dir)) ? $dir : $conf->val('Script', 'ScriptWebRoot');
	
	# Inform users
	print "     " . $lang->val('Messages','ReWriteDir') . " $dir\n";

	# Open Directory Handle
	opendir(DIR, "$dir/")
	 or die("Program was unable to open directory $dir: $!");

	# Read through directory listing
	while ($file = readdir(DIR))
	{
		# Skip parent and current
		next if ($file eq '.');
		next if ($file eq '..');

		# Skip based on file type
		next if (-l "$dir/$file");
		next if (-S "$dir/$file");
		next if (-p "$dir/$file");

		# Directory
		if (-d "$dir/$file")
		{
			# Add to list
			push(@locales, "$dir/$file");
		}

		# File
		elsif (-f "$dir/$file")

		{
			# Rewrite file if necessary
			ReWrite($conf, $lang, "$dir/$file");
		}
	}

	# Close directory
	closedir(DIR);


	# Call additional directories
	foreach $dir (sort(@locales))
	{
		# Handle subdirectory
		Search($conf, $lang, $dir);
	}

	# Insert spacer
	print "\n";
}
